import time
import json
from copy import deepcopy
from socrata.resource import Resource
from socrata.configs import Config
from socrata.http import noop, put, get, post


class ColumnChange(object):
    def __init__(self, field_name, attribute, output_schema):
        self._field_name = field_name
        self._attribute = attribute
        self._output_schema = output_schema

    def to(self, value):
        def change_fun(col):
            col[self._attribute] = value
            return col

        self._output_schema.column_changes.append((self._field_name, change_fun))

        return self._output_schema

class TransformChange(object):
    def __init__(self, field_name, output_schema):
        self._field_name = field_name
        self._output_schema = output_schema

    def to(self, value):
        def change_fun(col):
            col['transform']['transform_expr'] = value
            return col

        self._output_schema.column_changes.append((self._field_name, change_fun))

        return self._output_schema


class OutputSchema(Resource):
    """
        This is data as transformed from an InputSchema
    """

    def __init__(self, *args, **kwargs):
        super(OutputSchema, self).__init__(*args, **kwargs)
        self.column_changes = []
        self.column_additions = []
        self.column_deletions = []

    def build_config(self, uri, name, data_action):
        """
        Create a new ImportConfig from this OutputSchema. See the API
        docs for what an ImportConfig is and why they're useful
        """
        res = post(
            self.path(uri),
            auth = self.auth,
            data = json.dumps({
                'name': name,
                'data_action': data_action
            })
        )
        return Config(self.auth, res, None)

    def any_failed(self):
        self.wait_for_finish()
        """
        This is probably not the function you are looking for.

        This returns whether or not any transform in this output schema has failed. "Failed" in this
        case means an internal error (which is unexpected), not a data error (which is expected). This
        function will wait for processing to complete if it hasn't yet.

        For data errors:

            Tell whether or not there are data errors
                output_schema.any_errors()
            Get the count of data errors
                output_schema.attributes['error_count']
            Get the errors themselves
                output_schema.schema_errors(offset = 0, limit = 20)
        """
        return any([column['transform']['failed_at'] for column in self.attributes['output_columns']])


    def any_errors(self):
        self.wait_for_finish()
        """
        Whether or not any transform returned a data error in this schema. This
        function will wait for processing to complete if it hasn't yet.
        """

        return self.attributes['error_count'] > 0


    def wait_for_finish(self, progress = noop, timeout = None, sleeptime = 1):
        """
        Wait for this dataset to finish transforming and validating. Accepts a progress function
        and a timeout.
        """
        return self._wait_for_finish(
            is_finished = lambda m: m.attributes['completed_at'],
            is_failed = lambda m: m.any_failed(),
            progress = progress,
            timeout = timeout,
            sleeptime = sleeptime
        )

    def _munge_row(self, row):
        row = row['row']
        columns = sorted(self.attributes['output_columns'], key = lambda oc: oc['position'])
        field_names = [oc['field_name'] for oc in columns]
        return {k: v for k, v in zip(field_names, row)}


    def _get_rows(self, uri, offset, limit):
        resp = get(
            self.path(uri),
            params = {'limit': limit, 'offset': offset},
            auth = self.auth
        )

        rows = resp[1:]
        return [self._munge_row(row) for row in rows]

    def rows(self, uri, offset = 0, limit = 500):
        """
        Get the rows for this OutputSchema. Acceps `offset` and `limit` params
        for paging through the data.
        """
        return self._get_rows(uri, offset, limit)


    def schema_errors(self, uri, offset = 0, limit = 500):
        """
        Get the errors that resulted in transforming into this output schema.
        Accepts `offset` and `limit` params
        """
        return self._get_rows(uri, offset, limit)

    def schema_errors_csv(self):
        """
        Get the errors that results in transforming into this output schema
        as a CSV stream.

        Note that this returns a Reponse, where Reponse
        is a python requests Reponse object
        """
        return get(
            self.path(self.schema_errors_uri),
            auth = self.auth,
            headers = {'accept': 'text/csv', 'content-type': 'text/csv'},
            stream = True
        )

    def validate_row_id(self, uri, field_name):
        """
        Set the row id. Note you must call `validate_row_id` before doing this.

        Args:
        ```
            field_name (str): The column to validate as the row id
        ```

        Returns:
        ```
            boolean
        ```
        """
        output_column = [oc for oc in self.attributes['output_columns'] if oc['field_name'] == field_name]
        if len(output_column):
            [output_column] = output_column
            transform_id = output_column['transform']['id']

            return get(
                self.path(uri.format(transform_id = transform_id)),
                auth = self.auth
            )['valid']
        else:
            return False

    def set_row_id(self, field_name = None):
        """
        Set the row id. Note you must call `validate_row_id` before doing this.

        Args:
        ```
            field_name (str): The column to set as the row id
        ```

        Returns:
        ```
            OutputSchema
        ```

        Examples:
        ```python
        new_output_schema = output.set_row_id('the_id_column')
        ```
        """
        desired_schema = deepcopy(self.attributes['output_columns'])

        for oc in desired_schema:
            oc['is_primary_key'] = (oc['field_name'] == field_name)

        return self.parent.transform({'output_columns': desired_schema})


    def add_column(self, field_name, display_name, transform_expr, description = None):
        """
        Add a column

        Args:
        ```
            field_name (str): The column's field_name, must be unique
            display_name (str): The columns display name
            transform_expr (str): SoQL expression to evaluate and fill the column with data from
            description (str): Optional column description
        ```

        Returns:
        ```
            output_schema (OutputSchema): Returns self for easy chaining
        ```

        Examples:
        ```python
        new_output_schema = output
            # Add a new column, which is computed from the `celsius` column
            .add_column('fahrenheit', 'Degrees (Fahrenheit)', '(to_number(`celsius`) * (9 / 5)) + 32', 'the temperature in celsius')
            # Add a new column, which is computed from the `celsius` column
            .add_column('kelvin', 'Degrees (Kelvin)', '(to_number(`celsius`) + 273.15')
            .run()
        ```
        """
        position = len(self.attributes['output_columns']) + len(self.column_additions) - len(self.column_deletions)
        self.column_additions.append({
            'field_name': field_name,
            'display_name': display_name,
            'description': description,
            'position': position,
            'transform': {
                'transform_expr': transform_expr
            }
        })
        return self

    def drop_column(self, field_name):
        """
        Drop the column

        Args:
        ```
            field_name (str): The column to drop
        ```

        Returns:
        ```
            output_schema (OutputSchema): Returns self for easy chaining
        ```

        Examples:
        ```python
            new_output_schema = output
                .drop_column('foo')
                .run()
        ```
        """
        self.column_deletions.append(field_name)
        return self

    def change_column_metadata(self, field_name, attribute):
        """
        Change the column metadata. This returns a ColumnChange,
        which implements a `.to` function, which takes the new value to change to

        Args:
        ```
            field_name (str): The column to change
            attribute (str): The attribute of the column to change
        ```

        Returns:
        ```
            change (TransformChange): The transform change, which implements the `.to` function
        ```

        Examples:
        ```python
            new_output_schema = output
                # Change the field_name of date to the_date
                .change_column_metadata('date', 'field_name').to('the_date')
                # Change the description of the celsius column
                .change_column_metadata('celsius', 'description').to('the temperature in celsius')
                # Change the display name of the celsius column
                .change_column_metadata('celsius', 'display_name').to('Degrees (Celsius)')
                .run()
        ```
        """
        return ColumnChange(field_name, attribute, self)

    def change_column_transform(self, field_name):
        """
        Change the column transform. This returns a TransformChange,
        which implements a `.to` function, which takes a transform expression.

        Args:
        ```
            field_name (str): The column to change
        ```

        Returns:
        ```
            change (TransformChange): The transform change, which implements the `.to` function
        ```

        Examples:
        ```python
            new_output_schema = output
                .change_column_transform('the_date').to('to_fixed_timestamp(`date`)')
                # Make the celsius column all numbers
                .change_column_transform('celsius').to('to_number(`celsius`)')
                # Add a new column, which is computed from the `celsius` column
                .add_column('fahrenheit', 'Degrees (Fahrenheit)', '(to_number(`celsius`) * (9 / 5)) + 32', 'the temperature in celsius')
                .run()
        ```
        """
        return TransformChange(field_name, self)

    def run(self):
        """
        Run all adds, drops, and column changes.


        Returns:
        ```
            OutputSchema
        ```

        Examples:
        ```python
            new_output_schema = output
                # Change the field_name of date to the_date
                .change_column_metadata('date', 'field_name').to('the_date')
                # Change the description of the celsius column
                .change_column_metadata('celsius', 'description').to('the temperature in celsius')
                # Change the display name of the celsius column
                .change_column_metadata('celsius', 'display_name').to('Degrees (Celsius)')
                # Change the transform of the_date column to to_fixed_timestamp(`date`)
                .change_column_transform('the_date').to('to_fixed_timestamp(`date`)')
                # Make the celsius column all numbers
                .change_column_transform('celsius').to('to_number(`celsius`)')
                # Add a new column, which is computed from the `celsius` column
                .add_column('fahrenheit', 'Degrees (Fahrenheit)', '(to_number(`celsius`) * (9 / 5)) + 32', 'the temperature in celsius')
                .run()
        ```
        """
        columns = [
            c for c in deepcopy(self.attributes['output_columns'])
            if not (c['field_name'] in self.column_deletions)
        ] + self.column_additions

        for (to_change, change_fun) in self.column_changes:
            column = [
                column
                for column in columns
                if to_change == column['field_name']
            ]
            if not column:
                raise ValueError('Column `%s` does not exist' % to_change)
            else:
                [column] = column

            new_column = change_fun(column)

            # if there is an id, this means the column is not brand new
            if 'id' in column:
                new_column['initial_output_column_id'] = column['id']

            def replace_with(c):
                if c == column:
                    return new_column
                return c

            columns = [replace_with(c) for c in columns]

        columns = sorted(columns, key = lambda x: x['position'])

        for p, c in enumerate(columns):
            c['position'] = p + 1

        desired_schema = {
            'output_columns': columns
        }
        self.column_additions = []
        self.column_deletions = []
        self.column_changes = []
        return self.parent.transform(desired_schema)



import json
import io
import webbrowser
import types
from socrata.http import post, put, patch, get, noop
from socrata.resource import Resource, Collection, ChildResourceSpec
from socrata.input_schema import InputSchema
from socrata.builders.parse_options import ParseOptionBuilder
from socrata.lazy_pool import LazyThreadPoolExecutor
from threading import Lock

class Sources(Collection):
    def path(self):
        return 'https://{domain}/api/publishing/v1/source'.format(
            domain = self.auth.domain
        )

    def lookup(self, source_id):
        """
        Lookup a source

        Args:
        ```
            source_id (int): The id
        ```

        Returns:
        ```
            Source: Returns the new Source The Source resulting from this API call, or an error
        ```
        """
        return self._subresource(Source, get(
            self.path() + '/' + str(source_id),
            auth = self.auth
        ))

    def create_upload(self, filename):
        """
        Create a new source. Takes a `body` param, which must contain a `filename`
        of the file.

        Args:
        ```
            filename (str): The name of the file you are uploading
        ```

        Returns:
        ```
            Source: Returns the new Source
        ```

        Examples:
        ```python
            upload = revision.create_upload('foo.csv')
        ```

        """
        return self._subresource(Source, post(
            self.path(),
            auth = self.auth,
            data = json.dumps({
                'source_type' : {
                    'type': 'upload',
                    'filename': filename
                }
            })
        ))


class ChunkIterator(object):
    def __init__(self, filelike, chunk_size):
        self._filelike = filelike
        self.lock = Lock()
        self._chunk_size = chunk_size
        self.seq_num = 0
        self.byte_offset = 0

    def __iter__(self):
        return self

    def __next__(self):
        with self.lock:
            read = self._filelike.read(self._chunk_size)
            if not read:
                raise StopIteration

            this_seq = self.seq_num
            this_byte_offset = self.byte_offset
            self.seq_num = self.seq_num + 1
            self.byte_offset = self.byte_offset + len(read)
            return (this_seq, this_byte_offset, self.byte_offset, read)

    def next(self):
        return self.__next__()

class FileLikeGenerator(object):
    def __init__(self, gen):
        self.gen = gen
        self.done = False

    def read(self, how_much):
        if self.done:
            return None

        buf = []
        consumed = 0
        while consumed < how_much:
            try:
                chunk = next(self.gen)
                consumed += len(chunk)
                buf.append(chunk)
            except StopIteration:
                self.done = True
                break

        return b''.join(buf)


class Source(Resource, ParseOptionBuilder):
    def initiate(self, uri, content_type):
        return post(
            self.path(uri),
            auth = self.auth,
            data = json.dumps({ 'content_type': content_type })
        )

    def chunk(self, uri, seq_num, byte_offset, bytes):
        return post(
            self.path(uri).format(seq_num=seq_num, byte_offset=byte_offset),
            auth = self.auth,
            data = bytes,
            headers = { 'content-type': 'application/octet-stream' }
        )

    def commit(self, uri, seq_num, byte_offset):
        return post(
            self.path(uri).format(seq_num=seq_num, byte_offset=byte_offset),
            auth = self.auth
        )


    def _chunked_bytes(self, file_or_string_or_bytes_or_generator, content_type):

        if type(file_or_string_or_bytes_or_generator) is str:
            file_handle = io.StringIO(file_or_string_or_bytes_or_generator)
        elif type(file_or_string_or_bytes_or_generator) is bytes:
            file_handle = io.BytesIO(file_or_string_or_bytes_or_generator)
        elif isinstance(file_or_string_or_bytes_or_generator, types.GeneratorType):
            file_handle = FileLikeGenerator(file_or_string_or_bytes_or_generator)
        elif hasattr(file_or_string_or_bytes_or_generator, 'read'):
            file_handle = file_or_string_or_bytes_or_generator
        else:
            raise ValueError("The thing to upload must be a file, string, bytes, or generator which yields bytes")


        init = self.initiate(content_type)
        chunk_size = init['preferred_chunk_size']
        parallelism = init['preferred_upload_parallelism']

        def sendit(chunk):
            (seq_num, byte_offset, end_byte_offset, bytes) = chunk
            self.chunk(seq_num, byte_offset, bytes)
            return (seq_num, byte_offset, end_byte_offset)

        pool = LazyThreadPoolExecutor(parallelism)
        results = [r for r in pool.map(sendit, ChunkIterator(file_handle, chunk_size))]
        (seq_num, byte_offset, end_byte_offset) = sorted(results, key=lambda x: x[0])[-1]
        self.commit(seq_num, end_byte_offset)
        return self.show()


    """
    Uploads bytes into the source. Requires content_type argument
    be set correctly for the file handle. It's advised you don't
    use this method directly, instead use one of the csv, xls, xlsx,
    or tsv methods which will correctly set the content_type for you.
    """
    def bytes(self, uri, file_handle, content_type):
        # This is just for backwards compat
        self._chunked_bytes(file_handle, content_type)


    def load(self, uri = None):
        """
        Forces the source to load, if it's a view source.

        Returns:
        ```
            Source: Returns the new Source
        ```
        """
        return self._mutate(put(
            self.path(uri or (self.links['show'] + "/load")),
            auth = self.auth,
            data = {},
            headers = {
                'content-type': 'application/json'
            }
        ))

    def child_specs(self):
        return [
            ChildResourceSpec(
                self,
                'input_schemas',
                'input_schema_links',
                'schemas',
                InputSchema,
                'input_schema_id'
            )
        ]


    def blob(self, file_handle):
        """
        Uploads a Blob dataset. A blob is a file that will not be parsed as a data file,
        ie: an image, video, etc.


        Returns:
        ```
            Source: Returns the new Source
        ```

        Examples:
        ```python
            with open('my-blob.jpg', 'rb') as f:
                upload = upload.blob(f)
        ```

        """
        source = self
        if self.attributes['parse_options']['parse_source']:
            source = self.change_parse_option('parse_source').to(False).run()
        return source._chunked_bytes(file_handle, "application/octet-stream")


    def csv(self, file_handle):
        """
        Upload a CSV, returns the new input schema.

        Args:
        ```
            file_handle: The file handle, as returned by the python function `open()`
        ```

        Returns:
        ```
            Source: Returns the new Source
        ```

        Examples:
        ```python
            with open('my-file.csv', 'rb') as f:
                upload = upload.csv(f)
        ```
        """
        return self._chunked_bytes(file_handle, "text/csv")

    def xls(self, file_handle):
        """
        Upload an XLS, returns the new input schema

        Args:
        ```
            file_handle: The file handle, as returned by the python function `open()`
        ```

        Returns:
        ```
            Source: Returns the new Source
        ```

        Examples:
        ```python
            with open('my-file.xls', 'rb') as f:
                upload = upload.xls(f)
        ```
        """
        return self._chunked_bytes(file_handle, "application/vnd.ms-excel")

    def xlsx(self, file_handle):
        """
        Upload an XLSX, returns the new input schema.

        Args:
        ```
            file_handle: The file handle, as returned by the python function `open()`
        ```

        Returns:
        ```
            Source: Returns the new Source
        ```

        Examples:
        ```python
            with open('my-file.xlsx', 'rb') as f:
                upload = upload.xlsx(f)
        ```
        """
        return self._chunked_bytes(file_handle, "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet")

    def tsv(self, file_handle):
        """
        Upload a TSV, returns the new input schema.

        Args:
        ```
            file_handle: The file handle, as returned by the python function `open()`
        ```

        Returns:
        ```
            Source: Returns the new Source
        ```

        Examples:
        ```python
            with open('my-file.tsv', 'rb') as f:
                upload = upload.tsv(f)
        ```
        """
        return self._chunked_bytes(file_handle, "text/tab-separated-values")

    def shapefile(self, file_handle):
        """
        Upload a Shapefile, returns the new input schema.

        Args:
        ```
            file_handle: The file handle, as returned by the python function `open()`
        ```

        Returns:
        ```
            Source: Returns the new Source
        ```

        Examples:
        ```python
            with open('my-shapefile-archive.zip', 'rb') as f:
                upload = upload.shapefile(f)
        ```
        """
        return self._chunked_bytes(file_handle, "application/zip")

    def kml(self, file_handle):
        """
        Upload a KML file, returns the new input schema.

        Args:
        ```
            file_handle: The file handle, as returned by the python function `open()`
        ```

        Returns:
        ```
            Source: Returns the new Source
        ```

        Examples:
        ```python
            with open('my-kml-file.kml', 'rb') as f:
                upload = upload.kml(f)
        ```
        """
        return self._chunked_bytes(file_handle, "application/vnd.google-earth.kml+xml")


    def geojson(self, file_handle):
        """
        Upload a geojson file, returns the new input schema.

        Args:
        ```
            file_handle: The file handle, as returned by the python function `open()`
        ```

        Returns:
        ```
            Source: Returns the new Source
        ```

        Examples:
        ```python
            with open('my-geojson-file.geojson', 'rb') as f:
                upload = upload.geojson(f)
        ```
        """
        return self._chunked_bytes(file_handle, "application/vnd.geo+json")


    def df(self, dataframe):
        """
        Upload a pandas DataFrame, returns the new source.

        Args:
        ```
            file_handle: The file handle, as returned by the python function `open()`
        ```

        Returns:
        ```
            Source: Returns the new Source
        ```

        Examples:
        ```python
            import pandas
            df = pandas.read_csv('test/fixtures/simple.csv')
            upload = upload.df(df)
        ```
        """
        s = io.StringIO()
        dataframe.to_csv(s, index=False)
        return self._chunked_bytes(bytes(s.getvalue().encode()),"text/csv")

    def add_to_revision(self, uri, revision):
        """
        Associate this Source with the given revision.
        """
        return self._clone(patch(
            self.path(uri),
            auth = self.auth,
            data = json.dumps({
                'revision': {
                    'fourfour': revision.attributes['fourfour'],
                    'revision_seq': revision.attributes['revision_seq']
                }
            })
        ))

    def update(self, uri, body):
        return self._clone(post(
            self.path(uri),
            auth = self.auth,
            data = json.dumps(body)
        ))

    def show_input_schema(self, uri, input_schema_id):
        res = get(
            self.path(uri.format(input_schema_id = input_schema_id)),
            auth = self.auth
        )

        return self._subresource(InputSchema, res)

    def get_latest_input_schema(self):
        return max(self.input_schemas, key = lambda s: s.attributes['id'])

    def wait_for_finish(self, progress = noop, timeout = None, sleeptime = 1):
        """
        Wait for this dataset to finish transforming and validating. Accepts a progress function
        and a timeout.
        """
        return self._wait_for_finish(
            is_finished = lambda m: m.attributes['finished_at'],
            is_failed = lambda m: m.attributes['failed_at'],
            progress = progress,
            timeout = timeout,
            sleeptime = sleeptime
        )

    def ui_url(self):
        """
        This is the URL to the landing page in the UI for the sources

        Returns:
        ```
            url (str): URL you can paste into a browser to view the source UI
        ```
        """
        if not self.parent:
            raise NotImplementedError("UI for revisionless sources is not implemented (yet). Sorry!")

        revision = self.parent
        return revision.ui_url() +  '/sources/{source_id}/preview'.format(
            source_id = self.attributes['id']
        )

    def open_in_browser(self):
        """
        Open this source in your browser, this will open a window
        """
        webbrowser.open(self.ui_url(), new = 2)



from threading import Thread, Lock, Semaphore, Event
from queue import Queue


ONE_YEAR = 365 * 24 * 60 * 60

THREAD_DONE = object()


class LazyThreadPoolExecutor(object):
    def __init__(self, num_workers=1):
        self.num_workers = num_workers
        self.result_queue = Queue()
        self.thread_sem = Semaphore(num_workers)
        self._shutdown = Event()
        self.threads = []

    def map(self, predicate, iterable):
        self._shutdown.clear()
        self.iterable = ThreadSafeIterator(iterable)
        self._start_threads(predicate)
        return self._result_iterator()

    def shutdown(self, wait=True):
        self._shutdown.set()
        if wait:
            for t in self.threads:
                t.join()

    def _start_threads(self, predicate):
        for i in range(self.num_workers):
            t = Thread(
                name="LazyChild #{0}".format(i),
                target=self._make_worker(predicate)
            )
            t.daemon = True
            self.threads.append(t)
            t.start()

    def _make_worker(self, predicate):
        def _w():
            with self.thread_sem:
                for thing in self.iterable:
                    self.result_queue.put(predicate(thing))
                    if self._shutdown.is_set():
                        break
            self.result_queue.put(THREAD_DONE)
        return _w

    def _result_iterator(self):
        while 1:
            # Queue.get is not interruptable w/ ^C unless you specify a
            # timeout.
            # Hopefully one year is long enough...
            # See http://bugs.python.org/issue1360
            result = self.result_queue.get(True, ONE_YEAR)
            if result is not THREAD_DONE:
                yield result
            else:
                # if all threads have exited
                # sorry, this is kind of a gross way to use semaphores
                # break
                if self.thread_sem._value == self.num_workers:
                    break
                else:
                    continue



class ThreadSafeIterator(object):
    def __init__(self, it):
        self._it = iter(it)
        self.lock = Lock()

    def __iter__(self):
        return self

    def __next__(self):
        with self.lock:
            return self._it.__next__()



from socrata.operations.utils import get_filename
from socrata.operations.operation import Operation

class ConfiguredJob(Operation):
    def run(self, data, put_bytes, filename = None):
        filename = get_filename(data, filename)
        rev = self.properties['view'].revisions.create_using_config(
            self.properties['config']
        )
        source = rev.create_upload(filename)
        source = put_bytes(source)
        output_schema = source.get_latest_input_schema().get_latest_output_schema()
        output_schema = output_schema.wait_for_finish()
        job = rev.apply(output_schema = output_schema)
        return (rev, job)



from socrata.operations.utils import get_filename
from socrata.operations.operation import Operation

class Create(Operation):
    def run(self, data, put_bytes, filename = None):
        filename = get_filename(data, filename)
        rev = self.publish.new(self.properties['metadata'])
        source = rev.create_upload(filename)
        source = put_bytes(source)
        output_schema = source.get_latest_input_schema().get_latest_output_schema()
        output_schema = output_schema.wait_for_finish()
        return (rev, output_schema)
